# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Module for generating OpenMolcas input deck format.
"""

from typing import Union, TYPE_CHECKING

from qdk.chemistry.geometry import Geometry
from qdk.chemistry.solvers.util import (
    formatted_geometry_str,
    num_atoms_from_mol
)

from rdkit.Chem.rdmolops import GetFormalCharge

if TYPE_CHECKING:
    from rdkit.Chem.AllChem import Mol


GEOMETRY_LINE_SEP = "\n  "

OPENMOLCAS_TEMPLATE = """
&GATEWAY
Coord
  {num_atoms}
  {mol_name}
  {geometry}
Basis={basis}
Group={symmetry}

&SEWARD
{integral_keyword}

&SCF
Charge={charge}
Spin={spin}
{rasscf_section}
"""

OPENMOLCAS_TEMPLATE_CASSCF = """
&RASSCF
  Charge= {charge}
  Nactel  =  {num_active_el}
  Ras2  =  {num_active_orbitals}
  Ciroot  =  {nroot} {nroot} 1
"""

OPEN_MOLCAS_TEMPLATE_FCIDUMP = """
&RASSCF
  DMRG
  FCIDUMP
  TYPEINDEX
  Spin={spin}
  Charge={charge}

  RGINPUT
  nsweeps = 5
  max_bond_dimension = 500
  ENDRG
"""


def create_input_deck(
    mol: "Mol",
    mol_name: str = "",
    geometry: Union[str, Geometry] = None,
    charge: int = None,
    spin: str = 1,
    basis: str = "ANO-RCC-MB",
    symmetry: str = "C1",
    integral_keyword: str = "",
    method: str = "CASSCF",
    get_broombridge: bool = False,
    num_active_el: int = None,
    num_active_orbitals: int = 0,
    ci_root: int = 1
) -> str:
    """Convenience function for creating an OpenMolcas input formatted string.

    :param mol: Molecule object
    :type mol: Mol
    :param mol_name: Molecule name, defaults to ""
    :type mol_name: str, optional
    :param geometry: Molecule geometry, defaults to None
    :type geometry: Union[str, Geometry], optional
    :param charge: Charge, defaults to None
    :type charge: int, optional
    :param spin: Spin, defaults to 1
    :type spin: str, optional
    :param basis: Molecule basis, defaults to "ANO-RCC-MB"
    :type basis: str, optional
    :param symmetry: Molecule summetry, defaults to "C1"
    :type symmetry: str, optional
    :param integral_keyword: Integral keyword, e.g. "Cholesky", "DIRECT",
        "1CCD", defaults to ""
    :type integral_keyword: str, optional
    :param method: Method to use, e.g. "FCIDUMP" or "CASSCF", defaults to ""
    :type method: str, defaults to "CASSCF"
    :param get_broombridge: Get data for generating Broombridge file
    :type get_broombridge: bool, defaults to False
    :param num_active_el: Number of active electrons, defaults to None (will
        be calculated from molecule if unspecified)
    :type num_active_el: int, optional
    :param num_active_orbitals: number of orbitals in each symmetry for the
        RAS2 orbital subspace, defaults to 0
    :type num_active_orbitals: int, optional
    :param ci_root: CI root(s), see 
        https://molcas.gitlab.io/OpenMolcas/sphinx/users.guide/programs/rasscf.html#optional-keywords,
        defaults to 1 (ground state)
    :type ci_root: int, optional
    :return: OpenMolcas input string
    :rtype: str
    """
    formatted_geometry = formatted_geometry_str(
        mol=mol, 
        geometry=geometry, 
        line_sep=GEOMETRY_LINE_SEP
    )
    num_atoms = len(geometry) if isinstance(geometry, Geometry) \
        else num_atoms_from_mol(mol)
    charge = charge if charge is not None else GetFormalCharge(mol)
    num_active_el = num_active_el or num_active_orbitals // 2

    if method.upper() == "CASSCF" and not get_broombridge:
        rasscf_section = OPENMOLCAS_TEMPLATE_CASSCF.format(
            charge=charge,
            num_active_el=num_active_el,
            num_active_orbitals=num_active_orbitals,
            nroot=ci_root
        )

    elif get_broombridge is True:
        rasscf_section = OPEN_MOLCAS_TEMPLATE_FCIDUMP.format(
            spin=spin,
            charge=charge
        )

    else:
        rasscf_section = ""

    return OPENMOLCAS_TEMPLATE.format(
        num_atoms=num_atoms,
        mol_name=mol_name,
        geometry=formatted_geometry,
        basis=basis,
        symmetry=symmetry,
        integral_keyword=integral_keyword,
        charge=charge,
        spin=spin,
        rasscf_section=rasscf_section
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Module for Jupyter widget that displays JSMol editor
"""
from ..geometry import mol_to_xyz

from jupyter_jsmol import JsmolView
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rdkit.Chem.AllChem import Mol


class JsmolWidget(JsmolView):
    """Jupyter widget for JSMol molecular geometry visualization
    """
    def __init__(self, *args, **kwargs):
        self.default_info['color'] = 'white'
        super().__init__(*args, **kwargs)

    @classmethod
    def from_mol(cls, mol: "Mol", num_confs: int = 10, *args, **kwargs):
        """Generate JsmolWidget object from RDKit molecule

        :param mol: RDKit molecule to visualize
        :type mol: Mol
        :param num_confs: Number of conformers to generate, defaults to 10
        :type num_confs: int, optional
        :return: Widget instance
        :rtype: JsmolWidget
        """
        xyz = mol_to_xyz(mol, num_confs=num_confs)
        return cls.from_str(xyz)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Module that defines the Geometry and Element classes for keeping track of
XYZ data for molecules
"""

import re
from dataclasses import dataclass
from typing import List, Tuple, Union, TYPE_CHECKING

from qdk.chemistry.geometry.rdkit_convert import _mol_to_coordinates
from qdk.chemistry.geometry.xyz import (
    coordinates_to_xyz,
    element_coords_to_xyz
)

from rdkit.Chem import AllChem as Chem

if TYPE_CHECKING:
    from rdkit.Chem import Mol

FLOAT_PATTERN = "([+-]?[0-9]*[.][0-9]+)"
XYZ_PATTERN = f"(\\w)\\s+{FLOAT_PATTERN}\\s+{FLOAT_PATTERN}\\s+{FLOAT_PATTERN}"
ATOMS_PATTERN = "\\s*(?P<num_atoms>\\d+)\\s*\n"


@dataclass
class Element:
    """Class for keeping track of element XYZ
    coordinates for molecule geometry
    """
    name: str
    x: float
    y: float
    z: float

    @classmethod
    def from_tuple(cls, value: Tuple[str, float, float, float]):
        """Classmethod for constructing an Element instance from element coordinates

        :param value: Tuple with values (element name,
            x coordinate, y coordinate, z coordinate)
        :type value: Tuple[str, float, float, float]
        :return: Element instance
        :rtype: Element
        """
        name, x, y, z = value
        return cls(name=name, x=float(x), y=float(y), z=float(z))

    def to_xyz(self):
        return element_coords_to_xyz(
            name=self.name,
            x=self.x,
            y=self.y,
            z=self.z
        )


class Geometry(List[Element]):
    """Molecular geometry consisting of list of elements with
    XYZ coordinates
    """
    def __init__(self, *args, charge: Union[int, None] = None, **kwargs):
        self.charge = charge
        super().__init__(*args, **kwargs)

    @property
    def coordinates(self) -> Tuple[str, float, float, float]:
        """Get coordinates list of tuples (name, x, y, z)

        :return: List of coordinate tuples
        :rtype: Tuple[str, float, float, float]
        """
        for element in self:
            yield (element.name, element.x, element.y, element.z)

    @classmethod
    def from_mol(cls, mol: "Mol", num_confs: int = 10):
        """Classmethod for constructing a geometry instance from an RDKit
        molecule object

        :param mol: RDKit molecule object
        :type mol: Mol
        :param num_confs: Number of molecular conformers to generate, defaults
            to 10
        :type num_confs: int, optional
        """
        # This returns a plain list of tuples (element name, x, y, z)
        coordinates = _mol_to_coordinates(
            mol=mol,
            num_confs=num_confs
        )
        charge = Chem.GetFormalCharge(mol)

        return cls(
            [
                Element(name=name, x=x, y=y, z=z)
                for (name, x, y, z) in coordinates
            ],
            charge=charge
        )

    @classmethod
    def from_xyz(cls, xyz: str):
        """Generate geometry portion of NWChem file from XYZ data.
        The formatting of the .xyz file format is as follows:

            <number of atoms>
            comment line
            <element> <X> <Y> <Z>
            ...

        Source: https://en.wikipedia.org/wiki/XYZ_file_format.

        :param xyz: XYZ file format
        :type xyz: str
        """
        if "\n" not in xyz:
            return cls([])

        num_atoms = re.match(ATOMS_PATTERN, xyz)
        match = re.findall(XYZ_PATTERN, xyz)
        assert num_atoms is not None and len(match) == int(num_atoms.group("num_atoms")), f"Invalid XYZ file: \
number of atoms '{num_atoms.group('num_atoms')}' \
does not match number of elements found {len(match)}"
        return cls([Element.from_tuple(item) for item in match])

    def to_xyz(self, title: str = "unnamed"):
        """Convert geometry to XYZ-formatted string

        :param title: XYZ file title
        :type title: str
        :returns: XYZ-formatted string
        :rtype: str
        """
        return coordinates_to_xyz(
            number_of_atoms=len(self),
            charge=self.charge,
            coordinates=self.coordinates,
            title=title
        )


def format_geometry(geometry: Geometry, line_sep: str = "\n") -> str:
    """Format geometry into text format

    Example:
    <Element> <x> <y> <x>

    :param geometry: Geometry object
    :type geometry: Geometry
    :param line_sep: Line separator, defaults to "\n"
    :type line_sep: str
    :return: Geometry in text format, each element separated by given separator
    :rtype: str
    """
    return line_sep.join(el.to_xyz() for el in geometry)


def format_geometry_from_mol(mol: "Mol", line_sep: str = "\n") -> str:
    """Get geometry in text format from RDKit molecule object

    :param mol: RDKit molecule
    :type mol: Mol
    :param line_sep: Line separator, defaults to "\n"
    :type line_sep: str
    """
    g = Geometry.from_mol(mol)
    return format_geometry(geometry=g, line_sep=line_sep)


def format_geometry_from_xyz(xyz: str, line_sep: str = "\n") -> str:
    """Generate geometry text format from XYZ data.
    The formatting of the .xyz file format is as follows:

        <number of atoms>
        comment line
        <element> <X> <Y> <Z>
        ...

    Source: https://en.wikipedia.org/wiki/XYZ_file_format.

    :param xyz: XYZ file format
    :type xyz: str
    :param line_sep: Line separator, defaults to "\n"
    :type line_sep: str
    :return: Geometry in text format
    :rtype: str
    """
    g = Geometry.from_xyz(xyz)
    return format_geometry(geometry=g, line_sep=line_sep)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Module for converting coordinates to XYZ-formatted data

The formatting of the .xyz file format is as follows:

    <number of atoms>
    comment line
    <element> <X> <Y> <Z>
    ...

Source: https://en.wikipedia.org/wiki/XYZ_file_format.
"""

from typing import Iterable, Tuple


def element_coords_to_xyz(name: str, x: float, y: float, z: float) -> str:
    """Convert element to XYZ formatted line

    :param name: Element name
    :type name: str
    :param x: X coordinate
    :type x: float
    :param y: Y coordinate
    :type y: float
    :param z: Z coordinate
    :type z: float
    :return: XYZ formatted string
    :rtype: str
    """
    return f"{name} {x} {y} {z}"


def coordinates_to_xyz(
    number_of_atoms: int, 
    charge: int, 
    coordinates: Iterable[Tuple[str, float, float, float]],
    title: str = "unnamed"
) -> str:
    """Convert coordinates to XYZ file formatted string.

    :param number_of_atoms: Number of atoms in the conformer
    :type number_of_atoms: int
    :param charge: Charge of the conformer
    :type charge: int
    :param coordinates: List of tuples with values element name, x, y and z
        coordinates
    :type coordinates: List[Tuple[str, float, float, float]]
    :param title: XYZ file title
    :type title: str
    :return: XYZ file format
    :rtype: str
    """
    result = [
        f"{number_of_atoms}",
        title
    ]

    # Convert to <element> <x> <y> <z>
    result += [element_coords_to_xyz(*element) for element in coordinates]

    if charge:
        result.extend([
            "$set",
            f"chrg {charge}",
            "$end"
        ])

    return "\n".join(result)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import pytest
from unittest import mock

from qdk.chemistry.solvers.openmolcas import create_input_deck


@pytest.fixture()
def test_deck():
    return """
&GATEWAY
Coord
  3
  
  O 0.002 0.398 0.0
  H 0.762 -0.203 0.0
  H -0.764 -0.195 0.0
Basis=ANO-RCC-MB
Group=C1

&SEWARD


&SCF
Charge=0
Spin=1

&RASSCF
  DMRG
  FCIDUMP
  TYPEINDEX
  Spin=1
  Charge=0

  RGINPUT
  nsweeps = 5
  max_bond_dimension = 500
  ENDRG

"""


def test_openmolcas(geometry, h2o, test_deck):
    with mock.patch("qdk.chemistry.geometry.Geometry.from_mol") as _m:
        _m.return_value = geometry

        openmolcas_input = create_input_deck(
            charge=0,
            spin=1,
            get_broombridge=True,
            mol=h2o
        )

    assert openmolcas_input == test_deck



# Copyright (c) Microsoft Corporation.
# Copyright (c) 2018 Jensen Group
# Licensed under the MIT License.

"""
Module for generating rdkit molobj/smiles/molecular graph from free atoms
Implementation by Jan H. Jensen, based on the paper
    Yeonjoon Kim and Woo Youn Kim
    "Universal Structure Conversion Method for Organic Molecules: From Atomic Connectivity
    to Three-Dimensional Geometry"
    Bull. Korean Chem. Soc. 2015, Vol. 36, 1769-1777
    DOI: 10.1002/bkcs.10334
"""

import copy
import itertools
import sys

from rdkit.Chem import rdmolops
from rdkit.Chem import rdchem
try:
    from rdkit.Chem import rdEHTTools #requires RDKit 2019.9.1 or later
except ImportError:
    rdEHTTools = None

from collections import defaultdict

import numpy as np
import networkx as nx

from rdkit import Chem
from rdkit.Chem import AllChem, rdmolops

from qdk.chemistry._xyz2mol.util import *


def get_BO(AC, UA, DU, valences, UA_pairs, use_graph=True):
    """
    """
    BO = AC.copy()
    DU_save = []

    while DU_save != DU:
        for i, j in UA_pairs:
            BO[i, j] += 1
            BO[j, i] += 1

        BO_valence = list(BO.sum(axis=1))
        DU_save = copy.copy(DU)
        UA, DU = get_UA(valences, BO_valence)
        UA_pairs = get_UA_pairs(UA, AC, use_graph=use_graph)[0]

    return BO


def valences_not_too_large(BO, valences):
    """
    """
    number_of_bonds_list = BO.sum(axis=1)
    for valence, number_of_bonds in zip(valences, number_of_bonds_list):
        if number_of_bonds > valence:
            return False

    return True

def charge_is_OK(BO, AC, charge, DU, atomic_valence_electrons, atoms, valences,
                 allow_charged_fragments=True):
    # total charge
    Q = 0

    # charge fragment list
    q_list = []

    if allow_charged_fragments:

        BO_valences = list(BO.sum(axis=1))
        for i, atom in enumerate(atoms):
            q = get_atomic_charge(atom, atomic_valence_electrons[atom], BO_valences[i])
            Q += q
            if atom == 6:
                number_of_single_bonds_to_C = list(BO[i, :]).count(1)
                if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                    Q += 1
                    q = 2
                if number_of_single_bonds_to_C == 3 and Q + 1 < charge:
                    Q += 2
                    q = 1

            if q != 0:
                q_list.append(q)

    return (charge == Q)

def BO_is_OK(BO, AC, charge, DU, atomic_valence_electrons, atoms, valences,
    allow_charged_fragments=True):
    """
    Sanity of bond-orders
    args:
        BO -
        AC -
        charge -
        DU - 
    optional
        allow_charges_fragments - 
    returns:
        boolean - true of molecule is OK, false if not
    """

    if not valences_not_too_large(BO, valences):
        return False

    check_sum = (BO - AC).sum() == sum(DU)
    check_charge = charge_is_OK(BO, AC, charge, DU, atomic_valence_electrons, atoms, valences,
                                allow_charged_fragments)

    if check_charge and check_sum: 
        return True

    return False


def BO2mol(mol, BO_matrix, atoms, atomic_valence_electrons,
           mol_charge, allow_charged_fragments=True):
    """
    based on code written by Paolo Toscani
    From bond order, atoms, valence structure and total charge, generate an
    rdkit molecule.
    args:
        mol - rdkit molecule
        BO_matrix - bond order matrix of molecule
        atoms - list of integer atomic symbols
        atomic_valence_electrons -
        mol_charge - total charge of molecule
    optional:
        allow_charged_fragments - bool - allow charged fragments
    returns
        mol - updated rdkit molecule with bond connectivity
    """

    l = len(BO_matrix)
    l2 = len(atoms)
    BO_valences = list(BO_matrix.sum(axis=1))

    if (l != l2):
        raise RuntimeError('sizes of adjMat ({0:d}) and Atoms {1:d} differ'.format(l, l2))

    rwMol = Chem.RWMol(mol)

    bondTypeDict = {
        1: Chem.BondType.SINGLE,
        2: Chem.BondType.DOUBLE,
        3: Chem.BondType.TRIPLE
    }

    for i in range(l):
        for j in range(i + 1, l):
            bo = int(round(BO_matrix[i, j]))
            if (bo == 0):
                continue
            bt = bondTypeDict.get(bo, Chem.BondType.SINGLE)
            rwMol.AddBond(i, j, bt)

    mol = rwMol.GetMol()

    if allow_charged_fragments:
        mol = set_atomic_charges(
            mol,
            atoms,
            atomic_valence_electrons,
            BO_valences,
            BO_matrix,
            mol_charge)
    else:
        mol = set_atomic_radicals(mol, atoms, atomic_valence_electrons, BO_valences)

    return mol


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import pytest
import os

from rdkit.Chem import AllChem as Chem

from qdk.chemistry.geometry import Geometry, Element
from qdk.chemistry.molecule import Molecule


@pytest.fixture()
def h2o():
    mol = Chem.AddHs(Chem.MolFromSmiles("O"))
    return mol


@pytest.fixture()
def geometry(h2o):
    el = [
        Element("O", 0.002, 0.398, 0.0),
        Element("H", 0.762, -0.203, 0.0),
        Element("H", -0.764, -0.195, 0.0)
    ]
    return Geometry(el, charge=Chem.GetFormalCharge(h2o))


@pytest.fixture()
def h2o_xyz():
    path, _ = os.path.split(__file__)
    return os.path.join(path, "data", "h2o.xyz")


@pytest.fixture()
def caffeine_xyz():
    path, _ = os.path.split(__file__)
    return os.path.join(path, "data", "caffeine.xyz")


@pytest.fixture()
def h2o_nw():
    path, _ = os.path.split(__file__)
    return os.path.join(path, "data", "h2o.nw")


@pytest.fixture()
def h2o_molecule(h2o_xyz):
    return Molecule.from_xyz(h2o_xyz)


@pytest.fixture()
def caffeine(caffeine_xyz):
    return Molecule.from_xyz(caffeine_xyz)


@pytest.fixture()
def caffeine_output():
    path, _ = os.path.split(__file__)
    return os.path.join(path, "data", "caffeine.output")


@pytest.fixture()
def caffeine_nw():
    path, _ = os.path.split(__file__)
    return os.path.join(path, "data", "caffeine.nw")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
import rdkit

from qdk.chemistry.geometry.geometry import Geometry
from qdk.chemistry.molecule import Molecule


def test_from_xyz(caffeine_xyz):
    caffeine = Molecule.from_xyz(caffeine_xyz)
    assert caffeine.xyz() == caffeine._xyz
    Geometry.from_xyz(caffeine._xyz)
    formatted_xyz = "\n".join(" ".join(
        line.rstrip("00").replace("0 ", " ").split()
    ) for line in caffeine._xyz.split("\n"))
    assert caffeine.geometry.to_xyz("Caffeine") == formatted_xyz


def test_mol(caffeine):
    assert type(caffeine.mol) == rdkit.Chem.rdchem.Mol
    assert caffeine.smiles == "[H]Cn1c(=O)c2c(ncn2C)n(C)c1=O"
    assert caffeine.num_electrons == 102
    assert caffeine.atoms == [1, 6, 7, 8]
    assert caffeine.num_orbitals("STO-3G") == 80


def test_create_input(caffeine, tmp_path, h2o_molecule, h2o_nw):
    path = h2o_molecule.create_input(
        molecule_name="H2O",
        file_name="h2o.nw",
        solver="NWChem",
        base_path=tmp_path,
        num_active_orbitals=2
    )
    assert os.path.isfile(path)
    with open(path, "r") as f:
        data_gen = f.read()
    with open(h2o_nw, "r") as f:
        data_nw = f.read()
    assert data_gen == data_nw



# Copyright (c) Microsoft Corporation.
# Copyright (c) 2018 Jensen Group
# Licensed under the MIT License.

"""
Module for generating rdkit molobj/smiles/molecular graph from free atoms
Implementation by Jan H. Jensen, based on the paper
    Yeonjoon Kim and Woo Youn Kim
    "Universal Structure Conversion Method for Organic Molecules: From Atomic Connectivity
    to Three-Dimensional Geometry"
    Bull. Korean Chem. Soc. 2015, Vol. 36, 1769-1777
    DOI: 10.1002/bkcs.10334
"""

import logging
import sys
import itertools

from rdkit.Chem import rdmolops
from rdkit.Chem import rdchem

try:
    from rdkit.Chem import rdEHTTools  # requires RDKit 2019.9.1 or later
except ImportError:
    rdEHTTools = None


import numpy as np
import networkx as nx

from rdkit import Chem
from rdkit.Chem import AllChem, rdmolops

from qdk.chemistry._xyz2mol.util import *
from qdk.chemistry._xyz2mol.bo import *

_log = logging.getLogger(__name__)


def get_AC(mol, covalent_factor=1.3):
    """
    Generate adjacent matrix from atoms and coordinates.
    AC is a (num_atoms, num_atoms) matrix with 1 being covalent bond and 0 is not
    covalent_factor - 1.3 is an arbitrary factor
    args:
        mol - rdkit molobj with 3D conformer
    optional
        covalent_factor - increase covalent bond length threshold with facto
    returns:
        AC - adjacent matrix
    """

    # Calculate distance matrix
    dMat = Chem.Get3DDistanceMatrix(mol)

    pt = Chem.GetPeriodicTable()
    num_atoms = mol.GetNumAtoms()
    AC = np.zeros((num_atoms, num_atoms), dtype=int)

    for i in range(num_atoms):
        a_i = mol.GetAtomWithIdx(i)
        Rcov_i = pt.GetRcovalent(a_i.GetAtomicNum()) * covalent_factor
        for j in range(i + 1, num_atoms):
            a_j = mol.GetAtomWithIdx(j)
            Rcov_j = pt.GetRcovalent(a_j.GetAtomicNum()) * covalent_factor
            if dMat[i, j] <= Rcov_i + Rcov_j:
                AC[i, j] = 1
                AC[j, i] = 1

    return AC


def xyz2AC(atoms, xyz, charge, use_huckel=False):
    """
    atoms and coordinates to atom connectivity (AC)
    args:
        atoms - int atom types
        xyz - coordinates
        charge - molecule charge
    optional:
        use_huckel - Use Huckel method for atom connecitivty
    returns
        ac - atom connectivity matrix
        mol - rdkit molecule
    """

    if use_huckel:
        return xyz2AC_huckel(atoms, xyz, charge)
    else:
        return xyz2AC_vdW(atoms, xyz)


def xyz2AC_vdW(atoms, xyz):

    # Get mol template
    mol = get_proto_mol(atoms)

    # Set coordinates
    conf = Chem.Conformer(mol.GetNumAtoms())
    for i in range(mol.GetNumAtoms()):
        conf.SetAtomPosition(i, (xyz[i][0], xyz[i][1], xyz[i][2]))
    mol.AddConformer(conf)

    AC = get_AC(mol)

    return AC, mol


def xyz2AC_huckel(atomicNumList, xyz, charge):
    """
    args
        atomicNumList - atom type list
        xyz - coordinates
        charge - molecule charge
    returns
        ac - atom connectivity
        mol - rdkit molecule
    """
    mol = get_proto_mol(atomicNumList)

    conf = Chem.Conformer(mol.GetNumAtoms())
    for i in range(mol.GetNumAtoms()):
        conf.SetAtomPosition(i, (xyz[i][0], xyz[i][1], xyz[i][2]))
    mol.AddConformer(conf)

    num_atoms = len(atomicNumList)
    AC = np.zeros((num_atoms, num_atoms)).astype(int)

    mol_huckel = Chem.Mol(mol)
    mol_huckel.GetAtomWithIdx(0).SetFormalCharge(
        charge
    )  # mol charge arbitrarily added to 1st atom

    # Run in temporary path because rdkit generates empty "nul" and "run.out" output files
    with in_temp_path(cleanup=True) as path:
        passed, result = rdEHTTools.RunMol(mol_huckel)

    opop = result.GetReducedOverlapPopulationMatrix()
    tri = np.zeros((num_atoms, num_atoms))
    tri[
        np.tril(np.ones((num_atoms, num_atoms), dtype=bool))
    ] = opop  # lower triangular to square matrix
    for i in range(num_atoms):
        for j in range(i + 1, num_atoms):
            pair_pop = abs(tri[j, i])
            if pair_pop >= 0.15:  # arbitry cutoff for bond. May need adjustment
                AC[i, j] = 1
                AC[j, i] = 1

    return AC, mol


def AC2BO(AC, atoms, charge, allow_charged_fragments=True, use_graph=True):
    """
    implemenation of algorithm shown in Figure 2
    UA: unsaturated atoms
    DU: degree of unsaturation (u matrix in Figure)
    best_BO: Bcurr in Figure
    """
    # make a list of valences, e.g. for CO: [[4],[2,1]]
    valences_list_of_lists = []
    AC_valence = list(AC.sum(axis=1))

    for i, (atomicNum, valence) in enumerate(zip(atoms, AC_valence)):
        # valence can't be smaller than number of neighbourgs
        possible_valence = [x for x in atomic_valence[atomicNum] if x >= valence]
        if not possible_valence:
            _log.error(
                "Valence of atom",
                i,
                "is",
                valence,
                "which bigger than allowed max",
                max(atomic_valence[atomicNum]),
                ". Stopping",
            )
            sys.exit()
        valences_list_of_lists.append(possible_valence)

    # convert [[4],[2,1]] to [[4,2],[4,1]]
    valences_list = itertools.product(*valences_list_of_lists)

    best_BO = AC.copy()

    for valences in valences_list:

        UA, DU_from_AC = get_UA(valences, AC_valence)

        check_len = len(UA) == 0
        if check_len:
            check_bo = BO_is_OK(
                AC,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atoms,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )
        else:
            check_bo = None

        if check_len and check_bo:
            return AC, atomic_valence_electrons

        UA_pairs_list = get_UA_pairs(UA, AC, use_graph=use_graph)
        for UA_pairs in UA_pairs_list:
            BO = get_BO(AC, UA, DU_from_AC, valences, UA_pairs, use_graph=use_graph)
            status = BO_is_OK(
                BO,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atoms,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )
            charge_OK = charge_is_OK(
                BO,
                AC,
                charge,
                DU_from_AC,
                atomic_valence_electrons,
                atoms,
                valences,
                allow_charged_fragments=allow_charged_fragments,
            )

            if status:
                return BO, atomic_valence_electrons
            elif (
                BO.sum() >= best_BO.sum()
                and valences_not_too_large(BO, valences)
                and charge_OK
            ):
                best_BO = BO.copy()

    if not charge_OK:
        _log.debug("Warning: SMILES charge doesn't match input charge")
    return best_BO, atomic_valence_electrons


def AC2mol(mol, AC, atoms, charge, allow_charged_fragments=True, use_graph=True):
    """ """

    # convert AC matrix to bond order (BO) matrix
    BO, atomic_valence_electrons = AC2BO(
        AC,
        atoms,
        charge,
        allow_charged_fragments=allow_charged_fragments,
        use_graph=use_graph,
    )

    # add BO connectivity and charge info to mol object
    mol = BO2mol(
        mol,
        BO,
        atoms,
        atomic_valence_electrons,
        charge,
        allow_charged_fragments=allow_charged_fragments,
    )

    # BO2mol returns an arbitrary resonance form. Let's make the rest
    mols = rdchem.ResonanceMolSupplier(
        mol, Chem.UNCONSTRAINED_CATIONS, Chem.UNCONSTRAINED_ANIONS
    )
    mols = [mol for mol in mols]

    return mols
